"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const Path = require('path');
const command_1 = require('./command');
const object_1 = require('./object');
const error_1 = require('./error');
const util_1 = require('../util');
const COMMAND_NAME_REGEX = /^[\w\d]+(?:-[\w\d]+)*$/;
const HELP_OPTION_REGEX = /^(?:-[h?]|--help)$/;
/**
 * Clime command line interface.
 */
class CLI {
    constructor(
        /** Command entry name. */
        name, 
        /** Root directory of command modules. */
        root) {
        this.name = name;
        this.root = Path.resolve(root);
    }
    execute(argv, cwd = process.cwd()) {
        return __awaiter(this, void 0, void 0, function* () {
            let { sequence, args, path, possibleUnknownCommandName } = yield this.preProcessArguments(argv);
            let stats = yield util_1.safeStat(path);
            let description;
            if (stats && stats.isFile()) {
                let module = require(path);
                let TargetCommand = (module.default || module);
                if (TargetCommand.prototype instanceof command_1.Command) {
                    // This is a command module with an actual command.
                    if (!TargetCommand.decorated) {
                        throw new TypeError(`Command defined in module "${path}" does not seem to be intialized, make sure to decorate it with \`@command()\``);
                    }
                    TargetCommand.path = path;
                    TargetCommand.sequence = sequence;
                    let argsParser = new ArgsParser(TargetCommand);
                    let parsedArgs = yield argsParser.parse(sequence, args, cwd);
                    if (!parsedArgs) {
                        return yield command_1.HelpInfo.build({ TargetCommand });
                    }
                    let command = new TargetCommand();
                    let { args: commandArgs, extraArgs: commandExtraArgs, options: commandOptions, context } = parsedArgs;
                    return yield this.executeCommand(command, commandArgs, commandExtraArgs, commandOptions, context);
                }
                else if (Path.basename(path) === 'default.js') {
                    // This is a command module with only description and
                    // subcommands information.
                    path = Path.dirname(path);
                    description = TargetCommand.description;
                }
                else {
                    throw new TypeError(`Module "${path}" is expected to be a command`);
                }
            }
            let helpInfo = yield command_1.HelpInfo.build({ dir: path, description });
            if (possibleUnknownCommandName) {
                throw new UsageError(`Unknown subcommand "${possibleUnknownCommandName}"`, {
                    getHelp() {
                        return helpInfo;
                    }
                });
            }
            if (args.length && HELP_OPTION_REGEX.test(args[0])) {
                return helpInfo;
            }
            else {
                throw helpInfo;
            }
        });
    }
    /**
     * Mapping the command line arguments to a specific command file.
     */
    preProcessArguments(argv) {
        return __awaiter(this, void 0, void 0, function* () {
            let sequence = [this.name];
            let searchPath = this.root;
            let argsIndex = 0;
            let entryPath = Path.join(this.root, 'default.js');
            let targetPath = (yield util_1.safeStat(entryPath)) ? entryPath : searchPath;
            let possibleUnknownCommandName;
            let aliases;
            outer: for (let i = argsIndex; i < argv.length; i++) {
                let possibleCommandName = argv[i];
                if (!COMMAND_NAME_REGEX.test(possibleCommandName)) {
                    break;
                }
                let subcommands = yield CLI.getSubcommandDescriptors(searchPath);
                if (subcommands && subcommands.length) {
                    let metadata = new Map();
                    for (let subcommand of subcommands) {
                        metadata.set(subcommand.name, subcommand);
                        let aliases = subcommand.aliases || subcommand.alias && [subcommand.alias];
                        if (!aliases) {
                            continue;
                        }
                        for (let alias of aliases) {
                            metadata.set(alias, subcommand);
                        }
                    }
                    if (!metadata.has(possibleCommandName)) {
                        possibleUnknownCommandName = possibleCommandName;
                        break;
                    }
                    let descriptor = metadata.get(possibleCommandName);
                    // If `possibleCommandName` is an alias.
                    if (descriptor.name !== possibleCommandName) {
                        possibleCommandName = descriptor.name;
                    }
                    if (descriptor.filename) {
                        targetPath = Path.resolve(searchPath, descriptor.filename);
                        argsIndex = i + 1;
                        sequence.push(possibleCommandName);
                        continue outer;
                    }
                }
                searchPath = Path.join(searchPath, possibleCommandName);
                let possiblePaths = [
                    searchPath + '.js',
                    Path.join(searchPath, 'default.js'),
                    searchPath
                ];
                for (let possiblePath of possiblePaths) {
                    if (yield util_1.safeStat(possiblePath)) {
                        targetPath = possiblePath;
                        argsIndex = i + 1;
                        sequence.push(possibleCommandName);
                        continue outer;
                    }
                }
                possibleUnknownCommandName = possibleCommandName;
                // If a directory at path `searchPath` does not exist, stop searching.
                if (!(yield util_1.safeStat(searchPath))) {
                    break;
                }
            }
            return {
                sequence,
                args: argv.slice(argsIndex),
                path: targetPath,
                possibleUnknownCommandName
            };
        });
    }
    executeCommand(command, commandArgs, commandExtraArgs, commandOptions, context) {
        let executeMethodArgs = commandArgs.concat();
        if (commandExtraArgs) {
            executeMethodArgs.push(commandExtraArgs);
        }
        if (commandOptions) {
            executeMethodArgs.push(commandOptions);
        }
        if (context) {
            executeMethodArgs.push(context);
        }
        return command.execute(...executeMethodArgs);
    }
    getHelp() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield command_1.HelpInfo.build({
                dir: this.root
            });
        });
    }
    /** @internal */
    static getSubcommandDescriptors(dir) {
        return __awaiter(this, void 0, void 0, function* () {
            let path = Path.join(dir, 'default.js');
            if (!(yield util_1.safeStat(path))) {
                return undefined;
            }
            let commandModule = require(path);
            return commandModule.subcommands;
        });
    }
}
exports.CLI = CLI;
class ArgsParser {
    constructor(command) {
        this.helpProvider = command;
        this.paramDefinitions = command.paramDefinitions;
        this.requiredParamsNumber = command.requiredParamsNumber;
        this.paramsDefinition = command.paramsDefinition;
        this.optionsConstructor = command.optionsConstructor;
        this.optionDefinitions = command.optionDefinitions;
        this.contextConstructor = command.contextConstructor;
        if (this.optionDefinitions) {
            this.optionFlagMapping = new Map();
            this.optionDefinitionMap = new Map();
            for (let definition of this.optionDefinitions) {
                let { name, flag } = definition;
                this.optionDefinitionMap.set(name, definition);
                if (flag) {
                    this.optionFlagMapping.set(flag, name);
                }
            }
        }
    }
    parse(sequence, args, cwd) {
        return __awaiter(this, void 0, void 0, function* () {
            let that = this;
            let ContextConstructor = this.contextConstructor || command_1.Context;
            let context = new ContextConstructor({
                cwd,
                commands: sequence
            });
            args = args.concat();
            let OptionConstructor = this.optionsConstructor;
            let optionDefinitions = this.optionDefinitions;
            let optionDefinitionMap = this.optionDefinitionMap || new Map();
            let optionFlagMapping = this.optionFlagMapping || new Map();
            let requiredOptionSet;
            let paramDefinitions = this.paramDefinitions || [];
            let pendingParamDefinitions = paramDefinitions.concat();
            let paramsDefinition = this.paramsDefinition;
            let argsNumber = args.length;
            let commandArgs = [];
            let commandExtraArgs = paramsDefinition && [];
            let commandOptions;
            if (OptionConstructor) {
                commandOptions = new OptionConstructor();
                requiredOptionSet = new Set();
                for (let definition of optionDefinitions) {
                    let { name, key, type, required, validators, toggle, default: defaultValue } = definition;
                    if (required) {
                        requiredOptionSet.add(name);
                    }
                    if (toggle) {
                        commandOptions[key] = false;
                    }
                    else {
                        commandOptions[key] = typeof defaultValue === 'string' ?
                            yield castArgument(defaultValue, name, type, validators) :
                            defaultValue;
                    }
                }
            }
            while (args.length) {
                let arg = args.shift();
                if (arg === '-?' ||
                    (arg === '-h' && !optionFlagMapping.has('h')) ||
                    (arg === '--help' && !optionDefinitionMap.has('help'))) {
                    return undefined;
                }
                if (arg[0] === '-') {
                    if (arg[1] === '-') {
                        yield consumeToggleOrOption(arg.substr(2));
                    }
                    else {
                        yield consumeFlags(arg.substr(1));
                    }
                }
                else if (pendingParamDefinitions.length) {
                    let definition = pendingParamDefinitions.shift();
                    let casted = yield castArgument(arg, definition.name, definition.type, definition.validators);
                    commandArgs.push(casted);
                }
                else if (paramsDefinition) {
                    let casted = yield castArgument(arg, paramsDefinition.name, paramsDefinition.type, paramsDefinition.validators);
                    commandExtraArgs.push(casted);
                }
                else {
                    throw new UsageError(`Expecting ${paramDefinitions.length} parameter(s) at most but got ${argsNumber} instead`, this.helpProvider);
                }
            }
            {
                let expecting = this.requiredParamsNumber;
                let got = commandArgs.length;
                if (got < expecting) {
                    let missingArgNames = pendingParamDefinitions
                        .slice(0, expecting - got)
                        .map(definition => `\`${definition.name}\``);
                    throw new UsageError(`Expecting parameter(s) ${missingArgNames.join(', ')}`, this.helpProvider);
                }
            }
            {
                let missingOptionNames = requiredOptionSet && requiredOptionSet.keys();
                if (missingOptionNames && missingOptionNames.length) {
                    throw new UsageError(`Missing required option(s) \`${missingOptionNames.join('`, `')}\``, this.helpProvider);
                }
            }
            for (let definition of pendingParamDefinitions) {
                let defaultValue = definition.default;
                let value = typeof defaultValue === 'string' ?
                    yield castArgument(defaultValue, definition.name, definition.type, definition.validators) :
                    defaultValue;
                commandArgs.push(value);
            }
            if (paramsDefinition &&
                paramsDefinition.required &&
                !commandExtraArgs.length) {
                throw new UsageError(`Expecting at least one element for letiadic parameters \`${paramsDefinition.name}\``, this.helpProvider);
            }
            return {
                args: commandArgs,
                extraArgs: paramsDefinition && commandExtraArgs,
                options: commandOptions,
                context: this.contextConstructor ? context : undefined
            };
            function consumeFlags(flags) {
                return __awaiter(this, void 0, void 0, function* () {
                    for (let i = 0; i < flags.length; i++) {
                        let flag = flags[i];
                        if (!optionFlagMapping.has(flag)) {
                            throw new UsageError(`Unknown option flag "${flag}"`, that.helpProvider);
                        }
                        let name = optionFlagMapping.get(flag);
                        let definition = optionDefinitionMap.get(name);
                        if (definition.required) {
                            requiredOptionSet.delete(name);
                        }
                        if (definition.toggle) {
                            commandOptions[definition.key] = true;
                        }
                        else {
                            if (i !== flags.length - 1) {
                                throw new UsageError('Only the last flag in a sequence can refer to an option instead of a toggle', that.helpProvider);
                            }
                            yield consumeOption(definition);
                        }
                    }
                });
            }
            function consumeToggleOrOption(name) {
                return __awaiter(this, void 0, void 0, function* () {
                    if (!optionDefinitionMap.has(name)) {
                        throw new UsageError(`Unknown option \`${name}\``, that.helpProvider);
                    }
                    let definition = optionDefinitionMap.get(name);
                    if (definition.required) {
                        requiredOptionSet.delete(name);
                    }
                    if (definition.toggle) {
                        commandOptions[definition.key] = true;
                    }
                    else {
                        yield consumeOption(definition);
                    }
                });
            }
            function consumeOption(definition) {
                return __awaiter(this, void 0, void 0, function* () {
                    let { name, key, type, validators } = definition;
                    let arg = args.shift();
                    if (arg === undefined) {
                        throw new UsageError(`Expecting value for option \`${name}\``, that.helpProvider);
                    }
                    if (arg[0] === '-') {
                        throw new UsageError(`Expecting a value instead of an option or toggle "${arg}" for option \`${name}\``, that.helpProvider);
                    }
                    commandOptions[key] = yield castArgument(arg, name, type, validators);
                });
            }
            // TODO: support casting provider object.
            function castArgument(arg, name, type, validators) {
                return __awaiter(this, void 0, void 0, function* () {
                    let value;
                    switch (type) {
                        case String:
                            value = arg;
                            break;
                        case Number:
                            value = Number(arg);
                            if (isNaN(value)) {
                                throw new error_1.ExpectedError(`Value "${arg}" cannot be casted to number`);
                            }
                            break;
                        case Boolean:
                            if (/^(?:f|false)$/i.test(arg)) {
                                value = false;
                            }
                            else {
                                let n = Number(arg);
                                value = isNaN(n) ? true : Boolean(n);
                            }
                            break;
                        default:
                            if (object_1.isStringCastable(type)) {
                                value = yield type.cast(arg, context);
                            }
                            else {
                                throw new Error(`Type \`${type.name || type}\` cannot be casted from a string, see \`StringCastable\` interface for more information`);
                            }
                            break;
                    }
                    for (let validator of validators) {
                        if (validator instanceof RegExp) {
                            if (!validator.test(value)) {
                                throw new error_1.ExpectedError(`Invalid value for "${name}"`);
                            }
                        }
                        else {
                            validator.validate(arg, name);
                        }
                    }
                    return value;
                });
            }
        });
    }
}
class UsageError extends error_1.ExpectedError {
    constructor(message, helpProvider) {
        super(message);
        this.helpProvider = helpProvider;
    }
    print(stdout, stderr) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            _super("print").call(this, stdout, stderr);
            let help = yield this.helpProvider.getHelp();
            help.print(stdout, stderr);
        });
    }
}
exports.UsageError = UsageError;
//# sourceMappingURL=cli.js.map