"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const Path = require('path');
const FS = require('fs');
const Chalk = require('chalk');
const villa_1 = require('villa');
const cli_1 = require('../cli');
const util_1 = require('../../util');
class HelpInfo {
    constructor() {
        this.texts = [];
    }
    get text() {
        return this.texts.join('\n');
    }
    /** @internal */
    static build(options) {
        return __awaiter(this, void 0, void 0, function* () {
            let info = new HelpInfo();
            if (isHelpInfoBuildPathOptions(options)) {
                info.buildDescription(options.description);
                yield info.buildTextForSubCommands(options.dir);
            }
            else {
                let TargetCommand = options.TargetCommand;
                info.buildDescription(TargetCommand.description);
                info.buildTextsForParamsAndOptions(TargetCommand);
                let dir = Path.dirname(TargetCommand.path);
                if (Path.basename(TargetCommand.path) !== 'default.js') {
                    dir = Path.join(dir, Path.basename(TargetCommand.path, '.js'));
                }
                yield info.buildTextForSubCommands(dir);
            }
            return info;
        });
    }
    buildDescription(description) {
        if (description) {
            this.texts.push(`${util_1.indent(description)}\n`);
        }
    }
    buildTextsForParamsAndOptions(TargetCommand) {
        let paramDefinitions = TargetCommand.paramDefinitions;
        let paramsDefinition = TargetCommand.paramsDefinition;
        let parameterDescriptionRows = [];
        let parameterUsageTexts = [];
        if (paramDefinitions) {
            parameterUsageTexts = paramDefinitions.map(definition => {
                let { name, required, description, default: defaultValue } = definition;
                if (description) {
                    parameterDescriptionRows.push([
                        Chalk.bold(name),
                        description
                    ]);
                }
                return required ?
                    `<${name}>` :
                    `[${name}${defaultValue !== undefined ? '=' + defaultValue : ''}]`;
            });
        }
        else {
            parameterUsageTexts = [];
        }
        if (paramsDefinition) {
            let { name, required, description } = paramsDefinition;
            if (description) {
                parameterDescriptionRows.push([
                    Chalk.bold(name),
                    description
                ]);
            }
            parameterUsageTexts.push(required ?
                `<...${name}>` :
                `[...${name}]`);
        }
        let optionDefinitions = TargetCommand.optionDefinitions || [];
        let requiredOptionUsageItems = optionDefinitions
            .filter(definition => definition.required)
            .map(({ name, placeholder }) => `--${name} <${placeholder || name}>`);
        let usageLine = [
            Chalk.bold(TargetCommand.sequence.join(' ')),
            ...parameterUsageTexts,
            ...requiredOptionUsageItems
        ].join(' ');
        if (optionDefinitions.length > requiredOptionUsageItems.length) {
            usageLine += ' [...options]';
        }
        let usageContent = `\
  ${Chalk.green('USAGE')}\n
    ${usageLine}\n`;
        this.texts.push(usageContent);
        if (parameterDescriptionRows.length) {
            let paramsContent = `\
  ${Chalk.green('PARAMETERS')}\n
${util_1.buildTableOutput(parameterDescriptionRows, { indent: 4, spaces: ' - ' })}`;
            this.texts.push(paramsContent);
        }
        if (optionDefinitions.length) {
            let optionRows = optionDefinitions
                .map(definition => {
                let { name, key, flag, placeholder, toggle: isToggle, description } = definition;
                let triggerStr = flag ? `-${flag}, ` : '';
                triggerStr += `--${name}`;
                if (!isToggle) {
                    triggerStr += ` <${placeholder || key}>`;
                }
                return [
                    Chalk.bold(triggerStr),
                    description
                ];
            });
            this.texts.push(`\
  ${Chalk.green('OPTIONS')}\n
${util_1.buildTableOutput(optionRows, { indent: 4, spaces: ' - ' })}`);
        }
    }
    buildTextForSubCommands(dir) {
        return __awaiter(this, void 0, void 0, function* () {
            let rows;
            let subcommands = yield cli_1.CLI.getSubcommandDescriptors(dir);
            if (subcommands) {
                rows = subcommands.map(subcommand => {
                    let aliases = subcommand.aliases || subcommand.alias && [subcommand.alias];
                    let subcommandNamesStr = Chalk.bold(subcommand.name);
                    if (aliases) {
                        subcommandNamesStr += ` [${Chalk.dim(aliases.join(','))}]`;
                    }
                    return [
                        subcommandNamesStr,
                        subcommand.brief
                    ];
                });
            }
            else {
                let stats = yield util_1.safeStat(dir);
                if (!stats || !stats.isDirectory()) {
                    return;
                }
                let names = yield villa_1.call(FS.readdir, dir);
                let unfilteredRows = yield villa_1.map(names, (name) => __awaiter(this, void 0, void 0, function* () {
                    let path = Path.join(dir, name);
                    let stats = yield util_1.safeStat(path);
                    if (!stats) {
                        return undefined;
                    }
                    if (stats.isFile()) {
                        if (name === 'default.js' || Path.extname(path) !== '.js') {
                            return undefined;
                        }
                        name = Path.basename(name, '.js');
                    }
                    else {
                        path = Path.join(path, 'default.js');
                        stats = yield util_1.safeStat(path);
                    }
                    let description;
                    if (stats) {
                        let module = require(path);
                        let CommandClass = (module.default || module);
                        description = CommandClass && (CommandClass.brief || CommandClass.description);
                    }
                    return [
                        Chalk.bold(name),
                        description
                    ];
                }));
                rows = unfilteredRows.filter(row => !!row);
            }
            if (rows.length) {
                this.texts.push(`\
  ${Chalk.green('SUBCOMMANDS')}\n
${util_1.buildTableOutput(rows, { indent: 4, spaces: ' - ' })}`);
            }
        });
    }
    print(stdout, stderr) {
        stderr.write(`\n${this.text}\n`);
    }
}
exports.HelpInfo = HelpInfo;
function isHelpInfoBuildPathOptions(options) {
    return 'dir' in options;
}
//# sourceMappingURL=help.js.map